
/*
 * moblin-app-installer -- Moblin garage client application.
 * Copyright (c) 2009, Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms and conditions of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Author: Robert Staudinger <robertx.staudinger@intel.com>
 */

#include <glib/gi18n.h>
#include <gtk/gtk.h>

/* FIXME make conditional. */
#include <rpm/rpmlib.h>

#include "mai-model-priv.h"
#include "mai-package.h"
#include "mai-utils.h"
#include "config.h"

#define MAI_PACKAGE_ERROR (mai_package_error_quark ())

static GQuark
mai_package_error_quark (void)
{
  static GQuark _quark = 0;
  if (!_quark)
    _quark = g_quark_from_static_string ("mai-package-error");
  return _quark;
}

#define MAI_GOBJECT_PROPERTY_TRANSACTION(gobject_, statements_)               \
          {                                                                   \
            g_object_freeze_notify (G_OBJECT (gobject_));                     \
            statements_                                                       \
            g_object_thaw_notify (G_OBJECT (gobject_));                       \
          }

/*
 * MaiPackage
 */

G_DEFINE_TYPE (MaiPackage, mai_package, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE ((o), MAI_TYPE_PACKAGE, MaiPackagePrivate))

enum
{
  PROP_0,

  PROP_PACKAGE_ID,
  PROP_PACKAGE_NAME,
  PROP_NAME,
  PROP_VERSION,
  PROP_DESCRIPTION,
  PROP_LONG_DESCRIPTION,
  PROP_ICON,
  PROP_CATEGORIES,
  PROP_STATUS,
  PROP_PROGRESS,
  PROP_MESSAGE
};

typedef struct {
  /* Properties */
  char                 *package_id;
  gchar                *package_name;
  gchar                *name;
  gchar                *version;
  gchar                *description;
  gchar                *long_description;
  gchar                *icon;
  gchar               **categories;
  MaiPackageStatus      status;
  gint                  progress;
  /* Data */
  gchar                *message;
} MaiPackagePrivate;

static void
mai_package_set_name              (MaiPackage       *self,
                                   const gchar      *name);
static void
mai_package_set_version           (MaiPackage       *self,
                                   const gchar      *version);
static void
mai_package_set_description       (MaiPackage       *self,
                                   const gchar      *description);
static void
mai_package_set_long_description  (MaiPackage       *self,
                                  const gchar       *long_description);
static void
mai_package_set_icon              (MaiPackage       *self,
                                   const gchar      *icon);
static void
mai_package_set_categories        (MaiPackage       *self,
                                   const gchar     **categories);
static void
mai_package_set_progress          (MaiPackage       *self,
                                   gint              percentage);

static const gchar *
mai_package_status_to_text (MaiPackage const *self)
{
  MaiPackagePrivate *priv;

  g_return_val_if_fail (self, NULL);

  priv = GET_PRIVATE (self);

  switch (priv->status)
  {
  case MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE:
    return _("Queued for resolve");
  case MAI_PACKAGE_STATUS_RESOLVING:
    return _("Resolving ...");
  case MAI_PACKAGE_STATUS_AVAILABLE:
    return _("Available");
  case MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION:
    return _("Queued for installation");
  case MAI_PACKAGE_STATUS_INSTALLING:
    return _("Installing ...");
  case MAI_PACKAGE_STATUS_INSTALLED:
    return _("Installed");
  case MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL:
    return _("Queued for removal");
  case MAI_PACKAGE_STATUS_REMOVING:
    return _("Uninstalling ...");
  case MAI_PACKAGE_STATUS_ERROR:
    if (priv->message)
      return priv->message;
    else
      return _("Not found");
  }

  return NULL;
}

static void
_client_error_cb (PkClient        *client,
                  PkErrorCodeEnum  error_code,
                  const gchar     *details,
                  MaiPackage      *self)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  gchar             *message;

  MAI_DEBUG ("%s: error '%s' (%s)", priv->package_name,
                                    pk_error_enum_to_text (error_code),
                                    details);

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,
    if (details)
    {
      message = g_strdup_printf ("%s: %s",
                                 pk_error_enum_to_text (error_code),
                                 details);
    } else {
      message = g_strdup (pk_error_enum_to_text (error_code));
    }
    mai_package_set_message (self, message);
    mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
    g_free (message);
  );
}

static void
_client_message_cb (PkClient      *client,
                    PkMessageEnum  message,
                    gchar         *details,
                    MaiPackage    *self)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);

  MAI_DEBUG ("%s: '%s' '%s'", priv->package_name,
                              pk_message_enum_to_text (message),
                              details);
}

static void
_client_progress_cb (PkClient       *client,
                     guint           percentage,
                     guint           subpercentage,
                     guint           elapsed,
                     guint           remaining,
                     MaiPackage     *self)
{
  /* MAI_DEBUG ("%d", percentage); */

  mai_package_set_progress (self, CLAMP (percentage, 0, 100));
}

static void
_client_package_cb (PkClient       *client,
                    PkPackageObj   *package,
                    MaiPackage     *self)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  const gchar       *version = NULL;

{
  gchar *package_id = pk_package_id_to_string (package->id);
  /* MAI_DEBUG ("%s", package_id); */
  g_free (package_id);
}

  /* Search may return multiple results, we match on "package". */
  if (0 != g_strcmp0 (mai_package_get_package_name (self),
                      pk_package_id_get_name (package->id)))
  {
    return;
  }

  /* Skip if we already have already found a newer package. */
  version = pk_package_id_get_version (package->id);
  g_return_if_fail (version);

  if (priv->package_id &&
      priv->version &&
      rpmvercmp (version, priv->version) < 0)
  {
    g_debug ("    skipping '%s'", version);
    return;
  }

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,

    {
      gchar *package_id = pk_package_id_to_string (package->id);
      mai_package_set_package_id (self, package_id);
      g_free (package_id);
    }

    /* Fall back to info from package database. */

    if (NULL == mai_package_get_name (self))
      mai_package_set_name (self, pk_package_id_get_name (package->id));

    if (NULL == mai_package_get_description (self))
      mai_package_set_description (self, package->summary);

    /* Update package status. */
    if (package->info == PK_INFO_ENUM_INSTALLED)
      mai_package_set_status (self, MAI_PACKAGE_STATUS_INSTALLED);
    else if (package->info == PK_INFO_ENUM_AVAILABLE)
    {
      /* Set available if we have network connection. */
      PkControl *control = mai_model_get_control ();
      GError *error = NULL;
      PkNetworkEnum network_state = pk_control_get_network_state (control, &error);
      if (error)
      {
        mai_package_set_message (self, error->message);
        mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
        g_clear_error (&error);
      } else {
        g_debug ("    network: '%s'", pk_network_enum_to_text (network_state));
        if (network_state == PK_NETWORK_ENUM_ONLINE ||
            network_state == PK_NETWORK_ENUM_WIRED ||
            network_state == PK_NETWORK_ENUM_WIFI ||
            network_state == PK_NETWORK_ENUM_MOBILE)
        {
          mai_package_set_status (self, MAI_PACKAGE_STATUS_AVAILABLE);
        } else {
          gchar *message = g_strdup_printf ("Network is '%s'",
                                            pk_network_enum_to_text (network_state));
          mai_package_set_message (self, message);
          g_free (message);
          mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
        }
      }
    }
  );
}

static void
_client_allow_cancel_cb (PkClient   *client,
                         gboolean    allow_cancel,
                         MaiPackage *self)
{
  MAI_DEBUG ("%d", allow_cancel);
}

static void
_client_finished_cb (PkClient       *client,
                     PkExitEnum      exit,
                     guint           runtime,
                     MaiPackage     *self)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);
  GError            *error = NULL;

  /* MAI_DEBUG ("%s", pk_exit_enum_to_text (exit)); */

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,

    if (exit == PK_EXIT_ENUM_SUCCESS)
    {
      switch (priv->status)
      {
      case MAI_PACKAGE_STATUS_RESOLVING:
        /* Tried to resolve but not found. FIXME be nicer about this. */
        mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
        break;
      case MAI_PACKAGE_STATUS_AVAILABLE:
        /* Tried to resolve and found available. Good. */
        break;
      case MAI_PACKAGE_STATUS_INSTALLING:
        /* Package installed, resolve again to get consistent package-kit
         * status for all backends. */
        mai_package_resolve (self, &error);
        if (error)
        {
          g_warning ("%s : %s", G_STRLOC, error->message);
          g_clear_error (&error);
        }
        break;
      case MAI_PACKAGE_STATUS_INSTALLED:
        /* Tried to resolve and found already installed. Good. */
        break;
      case MAI_PACKAGE_STATUS_REMOVING:
        /* Package installed, resolve again to get consistent package-kit
         * status for all backends. */
        mai_package_resolve (self, &error);
        if (error)
        {
          g_warning ("%s : %s", G_STRLOC, error->message);
          g_clear_error (&error);
        }
        break;
      case MAI_PACKAGE_STATUS_QUEUED_FOR_RESOLVE:
      case MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION:
      case MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL:
      case MAI_PACKAGE_STATUS_ERROR:
        g_warning ("%s : Exit code ok but package '%s' in error status", G_STRLOC,
                                                                         priv->package_name);
        break;
      }
    } else {
      mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
      g_warning ("%s : Package '%s' exit '%s'", G_STRLOC,
                                                priv->package_name,
                                                pk_exit_enum_to_text (exit));
    }

    mai_package_set_progress (self, -1);
  );

  g_object_unref (client);
}

static void
_client_status_cb (PkClient     *client,
                   PkStatusEnum  status,
                   MaiPackage   *self)
{
  MaiPackagePrivate *priv = GET_PRIVATE (self);

  MAI_DEBUG ("%s: %s", priv->package_name, pk_status_enum_to_text (status));
}

static PkClient *
mai_package_create_client (MaiPackage *self)
{
  PkClient *client;

  client = pk_client_new ();

  /* Look up package. */
  g_signal_connect (client, "error-code",
                    G_CALLBACK (_client_error_cb), self);
  g_signal_connect (client, "message",
                    G_CALLBACK (_client_message_cb), self);
  g_signal_connect (client, "progress-changed",
                    G_CALLBACK (_client_progress_cb), self);
  g_signal_connect (client, "package",
                    G_CALLBACK (_client_package_cb), self);
  g_signal_connect (client, "finished",
                    G_CALLBACK (_client_finished_cb), self);
  g_signal_connect (client, "status-changed",
                    G_CALLBACK (_client_status_cb), self);
  g_signal_connect (client, "allow-cancel",
                    G_CALLBACK (_client_allow_cancel_cb), self);

  return client;
}

static void
_get_property (GObject    *object,
               guint       property_id,
               GValue     *value,
               GParamSpec *pspec)
{
  switch (property_id)
  {
  case PROP_PACKAGE_ID:
    g_value_set_string (value,
                        mai_package_get_package_id (MAI_PACKAGE (object)));
    break;
  case PROP_PACKAGE_NAME:
    g_value_set_string (value,
                        mai_package_get_package_name (MAI_PACKAGE (object)));
    break;
  case PROP_NAME:
    g_value_set_string (value,
                        mai_package_get_name (MAI_PACKAGE (object)));
    break;
  case PROP_VERSION:
    g_value_set_string (value,
                        mai_package_get_version (MAI_PACKAGE (object)));
    break;
  case PROP_DESCRIPTION:
    g_value_set_string (value,
                        mai_package_get_description (MAI_PACKAGE (object)));
    break;
  case PROP_LONG_DESCRIPTION:
    g_value_set_string (value,
                        mai_package_get_long_description (MAI_PACKAGE (object)));
    break;
  case PROP_ICON:
    g_value_set_string (value,
                        mai_package_get_icon (MAI_PACKAGE (object)));
    break;
  case PROP_CATEGORIES:
    g_value_set_pointer (value,
                         mai_package_get_categories (MAI_PACKAGE (object)));
    break;
  case PROP_STATUS:
    g_value_set_uint (value,
                      mai_package_get_status (MAI_PACKAGE (object)));
    break;
  case PROP_PROGRESS:
    g_value_set_int (value,
                     mai_package_get_progress (MAI_PACKAGE (object)));
    break;
  case PROP_MESSAGE:
    g_value_set_string (value,
                        mai_package_get_message (MAI_PACKAGE (object)));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_set_property (GObject      *object,
               guint         property_id,
               const GValue *value,
               GParamSpec   *pspec)
{
  MaiPackagePrivate *priv = GET_PRIVATE (object);

  switch (property_id)
  {
  case PROP_PACKAGE_ID:
    g_warn_if_fail (priv->package_id == NULL);
    priv->package_id = g_value_dup_string (value);
    break;
  case PROP_PACKAGE_NAME:
    priv->package_name = g_value_dup_string (value);
    break;
  case PROP_NAME:
    mai_package_set_name (MAI_PACKAGE (object),
                          g_value_get_string (value));
    break;
  case PROP_VERSION:
    mai_package_set_version (MAI_PACKAGE (object),
                             g_value_get_string (value));
    break;
  case PROP_DESCRIPTION:
    mai_package_set_description (MAI_PACKAGE (object),
                                 g_value_get_string (value));
    break;
  case PROP_LONG_DESCRIPTION:
    mai_package_set_long_description (MAI_PACKAGE (object),
                                      g_value_get_string (value));
    break;
  case PROP_ICON:
    mai_package_set_icon (MAI_PACKAGE (object),
                          g_value_get_string (value));
    break;
  case PROP_CATEGORIES:
    mai_package_set_categories (MAI_PACKAGE (object),
                                g_value_get_pointer (value));
    break;
  case PROP_STATUS:
    mai_package_set_status (MAI_PACKAGE (object),
                            g_value_get_uint (value));
    break;
  case PROP_PROGRESS:
    mai_package_set_progress (MAI_PACKAGE (object),
                              g_value_get_int (value));
    break;
  case PROP_MESSAGE:
    mai_package_set_message (MAI_PACKAGE (object),
                             g_value_get_string (value));
    break;
  default:
    G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
  }
}

static void
_dispose (GObject *object)
{
  MaiPackagePrivate *priv = GET_PRIVATE (object);

  if (priv->package_id)
    g_free (priv->package_id), priv->package_id = NULL;

  if (priv->package_name)
    g_free (priv->package_name), priv->package_name = NULL;

  if (priv->name)
    g_free (priv->name), priv->name = NULL;

  if (priv->version)
    g_free (priv->version), priv->version = NULL;

  if (priv->description)
    g_free (priv->description), priv->description = NULL;

  if (priv->long_description)
    g_free (priv->long_description), priv->long_description = NULL;

  if (priv->icon)
    g_free (priv->icon), priv->icon = NULL;

  if (priv->categories)
    g_strfreev (priv->categories), priv->categories = NULL;

  if (priv->message)
    g_free (priv->message), priv->message = NULL;

  G_OBJECT_CLASS (mai_package_parent_class)->dispose (object);
}

static GObject *
_constructor (GType                  type,
              guint                  n_properties,
              GObjectConstructParam *properties)
{
  MaiPackage *self = (MaiPackage *)
                            G_OBJECT_CLASS (mai_package_parent_class)
                              ->constructor (type, n_properties, properties);
  const gchar *icon;

  /* Icon */
  /* FIXME: check the icon exists, otherwise set default too. */
  if (NULL == (icon = mai_package_get_icon (self)))
  {
    /* No icon, use generic one. */
    GtkIconInfo *info = gtk_icon_theme_lookup_icon (gtk_icon_theme_get_default (),
                                                    "application-x-executable",
                                                    48,
                                                    GTK_ICON_LOOKUP_GENERIC_FALLBACK);
    if (info)
    {
      mai_package_set_icon (self, gtk_icon_info_get_filename (info));
      gtk_icon_info_free (info);
    }
  }

  return G_OBJECT (self);
}

static void
mai_package_class_init (MaiPackageClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GParamFlags   param_flags;

  g_type_class_add_private (klass, sizeof (MaiPackagePrivate));

  object_class->constructor = _constructor;
  object_class->get_property = _get_property;
  object_class->set_property = _set_property;
  object_class->dispose = _dispose;

  param_flags = G_PARAM_READWRITE | G_PARAM_CONSTRUCT;

  g_object_class_install_property (object_class,
                                   PROP_PACKAGE_ID,
                                   g_param_spec_string ("package-id", "Package ID",
                                                        "ID of package",
                                                        NULL, G_PARAM_READWRITE));
  g_object_class_install_property (object_class,
                                   PROP_PACKAGE_NAME,
                                   g_param_spec_string ("package-name", "Package name",
                                                        "Name of package",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_NAME,
                                   g_param_spec_string ("name", "Name",
                                                        "Application name",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_VERSION,
                                   g_param_spec_string ("version", "Version",
                                                        "Application version",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_DESCRIPTION,
                                   g_param_spec_string ("description", "Description",
                                                        "Application description",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_LONG_DESCRIPTION,
                                   g_param_spec_string ("long-description",
                                                        "Long description",
                                                        "Long package description",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_ICON,
                                   g_param_spec_string ("icon", "Icon",
                                                        "Application icon",
                                                        NULL, param_flags));
  g_object_class_install_property (object_class,
                                   PROP_CATEGORIES,
                                   g_param_spec_pointer ("categories", "Categories",
                                                         "Application categories",
                                                         param_flags));
  /* FIXME turn into an enum derived type. */
  g_object_class_install_property (object_class,
                                   PROP_STATUS,
                                   g_param_spec_uint ("status", "Status",
                                                      "Package status",
                                                      MAI_PACKAGE_STATUS_RESOLVING,
                                                      MAI_PACKAGE_STATUS_ERROR,
                                                      MAI_PACKAGE_STATUS_RESOLVING,
                                                      param_flags));
  g_object_class_install_property (object_class,
                                   PROP_PROGRESS,
                                   g_param_spec_int ("progress", "Progress",
                                                      "Transaction progress",
                                                      -1, 100, -1,
                                                      param_flags));
  g_object_class_install_property (object_class,
                                   PROP_MESSAGE,
                                   g_param_spec_string ("message",
                                                        "Message",
                                                        "Package status message",
                                                        NULL, param_flags));
}

static void
mai_package_init (MaiPackage *self)
{
}

MaiPackage *
mai_package_new (const gchar  *package,
                 const gchar  *name,
                 const gchar  *version,
                 const gchar  *description,
                 const gchar  *long_description,
                 const gchar  *icon,
                 const gchar **categories)
{
  MaiPackage *self;

  g_return_val_if_fail (package, NULL);

  g_debug ("%s() '%s'", __FUNCTION__, package);

  self = g_object_new (MAI_TYPE_PACKAGE,
                      "package-name",     package,
                      "name",             name,
                      "version",          version,
                      "description",      description,
                      "long-description", long_description,
                      "icon",             icon,
                      "categories",       categories,
                      NULL);

  return self;
}

gchar const *
mai_package_get_package_id (MaiPackage *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->package_id;
}

void
mai_package_set_package_id (MaiPackage    *self,
                            gchar const   *package_id)
{
  MaiPackagePrivate  *priv;
  gchar             **id;
  gboolean            do_notify;

  g_return_if_fail (self);
  g_return_if_fail (package_id);

  priv = GET_PRIVATE (self);
  id = g_strsplit (package_id, ";", -1);
  g_return_if_fail (0 == g_strcmp0 (priv->package_name, id[0]));

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,

    /* Package IDs may compare equal even when in fact they are not,
     * e.g. for installed/uninstalled package status. So we always update
     * but only notify when different. */
    do_notify = priv->package_id ?
                  !pk_package_id_equal_strings (package_id, priv->package_id) :
                  TRUE;

    if (priv->package_id)
    {
      g_free (priv->package_id);
      priv->package_id = NULL;
    }
    if (package_id)
    {
      priv->package_id = g_strdup (package_id);
    }

    if (do_notify)
      g_object_notify (G_OBJECT (self), "package-id");

    mai_package_set_version (self, id[1]);
  );

  g_strfreev (id);
}

const gchar *
mai_package_get_package_name (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->package_name;
}

const gchar *
mai_package_get_name (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->name;
}

static void
mai_package_set_name (MaiPackage  *self,
                      const gchar *name)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (name);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (name, priv->name))
  {
    g_free (priv->name);
    priv->name = g_strdup (name);
    g_object_notify (G_OBJECT (self), "name");
  }
}

const gchar *
mai_package_get_version (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->version;
}

static void
mai_package_set_version (MaiPackage   *self,
                         const gchar  *version)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (version, priv->version))
  {
    g_free (priv->version);
    priv->version = g_strdup (version);
    g_object_notify (G_OBJECT (self), "version");
  }
}

const gchar *
mai_package_get_description (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->description;
}

static void
mai_package_set_description (MaiPackage   *self,
                             const gchar  *description)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (description);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (description, priv->description))
  {
    g_free (priv->description);
    priv->description = g_strdup (description);
    g_object_notify (G_OBJECT (self), "description");
  }
}

const gchar *
mai_package_get_long_description (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->long_description;
}

static void
mai_package_set_long_description (MaiPackage   *self,
                                  const gchar  *long_description)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (long_description);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (long_description, priv->long_description))
  {
    g_free (priv->long_description);
    priv->long_description = g_strdup (long_description);
    g_object_notify (G_OBJECT (self), "long-description");
  }
}

const gchar *
mai_package_get_icon (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->icon;
}

static void
mai_package_set_icon (MaiPackage  *self,
                      const gchar *icon)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);
  if (0 != g_strcmp0 (icon, priv->icon))
  {
    g_free (priv->icon);
    priv->icon = g_strdup (icon);
    g_object_notify (G_OBJECT (self), "icon");
  }
}

MaiPackageStatus
mai_package_get_status (MaiPackage const *self)
{
  g_return_val_if_fail (self, MAI_PACKAGE_STATUS_ERROR);

  return GET_PRIVATE (self)->status;
}

void
mai_package_set_status (MaiPackage       *self,
                        MaiPackageStatus  status)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_return_if_fail (status >= MAI_PACKAGE_STATUS_RESOLVING);
  g_return_if_fail (status <= MAI_PACKAGE_STATUS_ERROR);

  priv = GET_PRIVATE (self);
  if (status != priv->status)
  {
    priv->status = status;
    g_object_notify (G_OBJECT (self), "status");
  }
}

const gchar **
mai_package_get_categories (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return (const gchar **) GET_PRIVATE (self)->categories;
}

static void
mai_package_set_categories (MaiPackage   *self,
                            const gchar **categories)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);

  g_strfreev (priv->categories);
  priv->categories = g_strdupv ((gchar **) categories);
  g_object_notify (G_OBJECT (self), "categories");
}

const gchar *
mai_package_get_message (MaiPackage const *self)
{
  g_return_val_if_fail (self, NULL);

  return GET_PRIVATE (self)->message;
}

void
mai_package_set_message (MaiPackage   *self,
                         const gchar  *message)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);

  priv = GET_PRIVATE (self);

  if (0 != g_strcmp0 (priv->message, message))
  {
    g_free (priv->message);
    priv->message = g_strdup (message);
    g_object_notify (G_OBJECT (self), "message");
  }
}

gint
mai_package_get_progress (MaiPackage const *self)
{
  g_return_val_if_fail (self, 0);

  return GET_PRIVATE (self)->progress;
}

static void
mai_package_set_progress (MaiPackage  *self,
                          gint         percentage)
{
  MaiPackagePrivate *priv;

  g_return_if_fail (self);
  g_warn_if_fail (percentage >= -1);
  g_warn_if_fail (percentage <= 100);

  priv = GET_PRIVATE (self);
  percentage = CLAMP (percentage, -1, 100);
  if (percentage != priv->progress)
  {
    priv->progress = percentage;
    g_object_notify (G_OBJECT (self), "progress");
  }
}

/*
 * returns > 0 if self is newer
 *           0 if equal
 *         < 0 if version is newer
 */
gint
mai_package_compare_version (MaiPackage const  *self,
                             const gchar       *version)
{
  MaiPackagePrivate *priv;

  g_return_val_if_fail (self, 0);

  priv = GET_PRIVATE (self);
  g_return_val_if_fail (priv->version && version, 0);

  if (NULL == priv->package_id)
    return -1;

  /* FIXME support deb. */
  return rpmvercmp (priv->version, version);
}

gboolean
mai_package_resolve (MaiPackage  *self,
                     GError     **error)
{
  MaiPackagePrivate *priv;
  PkClient          *client;
  gboolean           ret = FALSE;

  g_return_val_if_fail (self, FALSE);

  priv = GET_PRIVATE (self);

  /* MAI_DEBUG ("'%s'", priv->package_name); */

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,
    mai_package_set_status (self, MAI_PACKAGE_STATUS_RESOLVING);
    mai_package_set_progress (self, 0);
  );

  client = mai_package_create_client (self);
  ret = pk_client_search_name (client, PK_FILTER_ENUM_NONE,
                               priv->package_name, error);
  if (!ret)
  {
    MAI_GOBJECT_PROPERTY_TRANSACTION (self,
      mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
      mai_package_set_progress (self, -1);
    );
  }

  return ret;
}

gboolean
mai_package_install (MaiPackage  *self,
                     GError     **error)
{
  MaiPackagePrivate *priv;
  PkClient          *client;
  gchar const       *package_ids[2] = { 0, };
  gboolean           ret = FALSE;

  g_return_val_if_fail (MAI_IS_PACKAGE (self), FALSE);

  priv = GET_PRIVATE (self);
  g_return_val_if_fail (priv->package_id, FALSE);

  if (priv->status != MAI_PACKAGE_STATUS_QUEUED_FOR_INSTALLATION)
  {
    if (error)
      *error = g_error_new (MAI_PACKAGE_ERROR,
                            MAI_PACKAGE_ERROR_OPERATION_UNAVAILABLE,
                            "Package status is '%s', and therefore can not be installed",
                            mai_package_status_to_text (self));
    return FALSE;
  }

  package_ids[0] = priv->package_id;
  /* MAI_DEBUG ("'%s'", package_ids[0]); */

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,
    mai_package_set_status (self, MAI_PACKAGE_STATUS_INSTALLING);
    mai_package_set_progress (self, 0);
  );

  client = mai_package_create_client (self);
#ifdef HAVE_PACKAGEKIT_GLIB_0_5
  ret = pk_client_install_packages (client, TRUE, (gchar **) package_ids, error);
#else
  ret = pk_client_install_packages (client, (gchar **) package_ids, error);
#endif
  if (!ret)
  {
    MAI_GOBJECT_PROPERTY_TRANSACTION (self,
      mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
      mai_package_set_progress (self, -1);
    );
  }

  return ret;
}

gboolean
mai_package_remove (MaiPackage  *self,
                    GError     **error)
{
  MaiPackagePrivate *priv;
  PkClient          *client;
  gchar const       *package_ids[2] = { 0, };
  gboolean           ret = FALSE;

  g_return_val_if_fail (MAI_IS_PACKAGE (self), FALSE);

  priv = GET_PRIVATE (self);
  g_return_val_if_fail (priv->package_id, FALSE);

  if (priv->status != MAI_PACKAGE_STATUS_QUEUED_FOR_REMOVAL)
  {
    if (error)
      *error = g_error_new (MAI_PACKAGE_ERROR,
                            MAI_PACKAGE_ERROR_OPERATION_UNAVAILABLE,
                            "Package status is '%s', and therefore can not be uninstalled",
                            mai_package_status_to_text (self));
    return FALSE;
  }

  package_ids[0] = priv->package_id;
  /* MAI_DEBUG ("'%s'", package_ids[0]); */

  MAI_GOBJECT_PROPERTY_TRANSACTION (self,
    mai_package_set_status (self, MAI_PACKAGE_STATUS_REMOVING);
    mai_package_set_progress (self, 0);
  );

  client = mai_package_create_client (self);
  ret = pk_client_remove_packages (client, (gchar **) package_ids, TRUE, TRUE,
                                   error);

  if (!ret)
  {
    MAI_GOBJECT_PROPERTY_TRANSACTION (self,
      mai_package_set_status (self, MAI_PACKAGE_STATUS_ERROR);
      mai_package_set_progress (self, -1);
    );
  }

  return ret;
}

